/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <v/gtkvcheckmenuitem.h>
#include <v/gtkvlayout.h>

static void     gtk_vlayout_class_init         (GtkVLayoutClass *class);
static void     gtk_vlayout_init               (GtkVLayout      *layout);

static void     gtk_vlayout_map                (GtkWidget      *widget);
static void     gtk_vlayout_unmap              (GtkWidget      *widget);
static void     gtk_vlayout_size_request       (GtkWidget      *widget,
		 			       GtkRequisition *requisition);
static void     gtk_vlayout_size_allocate      (GtkWidget      *widget,
					       GtkAllocation  *allocation);
static void     gtk_vlayout_draw               (GtkWidget      *widget, 
					       GdkRectangle   *area);
static gint	gtk_vlayout_expose	       (GtkWidget      *widget,
						GdkEventExpose *event);
static GtkVLayoutChild *find_widget(GtkVLayout *vlayout, GtkWidget *widget);


guint
gtk_vlayout_get_type (void)
{
    static GtkType vlayout_type = 0;

    if (!vlayout_type) {
      GtkTypeInfo vlayout_info =
      {
	"GtkVLayout",
	sizeof (GtkVLayout),
	sizeof (GtkVLayoutClass),
	(GtkClassInitFunc) gtk_vlayout_class_init,
	(GtkObjectInitFunc) gtk_vlayout_init,
	(GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL,
      };

      vlayout_type = gtk_type_unique (GTK_TYPE_CONTAINER, &vlayout_info);
    }

    return vlayout_type;
}

static void
gtk_vlayout_class_init (GtkVLayoutClass *class)
{
    GtkWidgetClass *widget_class;

    widget_class = (GtkWidgetClass*) class;

    widget_class->map = gtk_vlayout_map;
    widget_class->unmap = gtk_vlayout_unmap;
    widget_class->draw = gtk_vlayout_draw;
    widget_class->expose_event = gtk_vlayout_expose;
    widget_class->size_request = gtk_vlayout_size_request;
    widget_class->size_allocate = gtk_vlayout_size_allocate;
}

static void     
gtk_vlayout_init (GtkVLayout *vlayout)
{
    GTK_WIDGET_SET_FLAGS(vlayout, GTK_NO_WINDOW);

    vlayout->children = NULL;
    vlayout->spacing = 0;
}

GtkWidget *
gtk_vlayout_new (gint spacing)
{
    GtkVLayout *vlayout;

    vlayout = gtk_type_new (gtk_vlayout_get_type ());
    vlayout->spacing = spacing;

    return GTK_WIDGET (vlayout);
}

void 
gtk_vlayout_add (GtkVLayout	*vlayout, 
		 GtkWidget	*child,
		 GtkWidget	*left,
		 GtkWidget	*above)
{
    GtkVLayoutChild *child_info;

    g_return_if_fail (vlayout != NULL);
    g_return_if_fail (GTK_IS_VLAYOUT (vlayout));
    g_return_if_fail (child != NULL);
    g_return_if_fail (child->parent == NULL);

    child_info = g_new (GtkVLayoutChild, 1);
    child_info->widget = child;
    child_info->right = 0;
    child_info->bottom = 0;
    child_info->left = left ? find_widget(vlayout, left) : NULL;
    child_info->above = above ? find_widget(vlayout, above) : NULL;

    vlayout->children = g_list_append(vlayout->children, child_info);

    gtk_widget_set_parent (child, GTK_WIDGET(vlayout));

    if (GTK_WIDGET_VISIBLE (GTK_WIDGET(vlayout))) {
	if (GTK_WIDGET_REALIZED (GTK_WIDGET (vlayout)) &&
		!GTK_WIDGET_REALIZED (child))
	    gtk_widget_realize (child);

	if (GTK_WIDGET_MAPPED (GTK_WIDGET (vlayout)) &&
		!GTK_WIDGET_MAPPED (child))
	    gtk_widget_map (child);
    }

    if (GTK_WIDGET_VISIBLE (child) && GTK_WIDGET_VISIBLE (vlayout))
	gtk_widget_queue_resize (child);
}

static GtkVLayoutChild *
find_widget(GtkVLayout *vlayout, GtkWidget *widget)
{
    GtkVLayoutChild *child;
    GList *children;

    for (children = vlayout->children; children; children = children->next) {
	child = children->data;
	if (widget == child->widget)
	    return child;
    }
    return NULL;
}

static void
gtk_vlayout_map (GtkWidget *widget)
{
    GtkVLayout *vlayout;
    GtkVLayoutChild *child;
    GList *children;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_VLAYOUT (widget));

    vlayout = GTK_VLAYOUT (widget);
    GTK_WIDGET_SET_FLAGS (vlayout, GTK_MAPPED);

    for (children = vlayout->children; children; children = children->next) {
	child = children->data;

	if (GTK_WIDGET_VISIBLE (child->widget) &&
		!GTK_WIDGET_MAPPED (child->widget))
	    gtk_widget_map (child->widget);
    }
}

static void
gtk_vlayout_unmap (GtkWidget *widget)
{
    GtkVLayout *vlayout;
    GtkVLayoutChild *child;
    GList *children;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_VLAYOUT (widget));

    vlayout = GTK_VLAYOUT (widget);
    GTK_WIDGET_UNSET_FLAGS (vlayout, GTK_MAPPED);

    for (children = vlayout->children; children; children = children->next) {
	child = children->data;

	if (GTK_WIDGET_VISIBLE (child->widget) &&
		GTK_WIDGET_MAPPED (child->widget))
	    gtk_widget_unmap (child->widget);
    }
}

static void
gtk_vlayout_size_request (GtkWidget      *widget,
			   GtkRequisition *requisition)
{
    GtkVLayout *vlayout;
    GList *children;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_VLAYOUT(widget));
    g_return_if_fail (requisition != NULL);

    vlayout = GTK_VLAYOUT (widget);
    requisition->width = 0;
    requisition->height = 0;

    for (children = vlayout->children; children; children = children->next) {
	GtkVLayoutChild *child = children->data;
        GtkRequisition child_requisition;

	gtk_widget_size_request (child->widget, &child_requisition);
	child->right = child_requisition.width;
	child->bottom = child_requisition.height;
	if (child->left) child->right += child->left->right + vlayout->spacing;
	if (child->above) child->bottom += child->above->bottom + vlayout->spacing;

	requisition->width = MAX (requisition->width, child->right);
	requisition->height = MAX (requisition->height, child->bottom);
    }

    requisition->width += GTK_CONTAINER (vlayout)->border_width * 2;
    requisition->height += GTK_CONTAINER (vlayout)->border_width * 2;
}

static void
gtk_vlayout_size_allocate (GtkWidget      *widget,
			   GtkAllocation  *allocation)
{
    GtkVLayout *vlayout;
    GList *children;
    GtkAllocation child_allocation;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_VLAYOUT(widget));
    g_return_if_fail (allocation != NULL);

    vlayout = GTK_VLAYOUT (widget);
    widget->allocation = *allocation;

    for (children = vlayout->children; children; children = children->next) {
	GtkVLayoutChild *child = children->data;
        GtkRequisition child_requisition;

	child_allocation.x = child->left ? child->left->right + vlayout->spacing
		 : allocation->x + GTK_CONTAINER (vlayout)->border_width;
	child_allocation.y = child->above ? child->above->bottom + vlayout->spacing
		 : allocation->y + GTK_CONTAINER (vlayout)->border_width;
	gtk_widget_size_request (child->widget, &child_requisition);
	child_allocation.width = child_requisition.width;
	child_allocation.height = child_requisition.height;

	gtk_widget_size_allocate (child->widget, &child_allocation);

	child->right = child_allocation.x + child_allocation.width;
	child->bottom = child_allocation.y + child_allocation.height;
    }
}

static void
gtk_vlayout_draw (GtkWidget      *widget, 
		   GdkRectangle   *area)
{
    GtkVLayout *vlayout;
    GtkVLayoutChild *child;
    GList *children;
    GdkRectangle child_area;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_VLAYOUT(widget));

    if (GTK_WIDGET_DRAWABLE (widget)) {
	vlayout = GTK_VLAYOUT (widget);

	for (children = vlayout->children; children; children = children->next) {
	    child = children->data;

	    if (gtk_widget_intersect (child->widget, area, &child_area))
		gtk_widget_draw (child->widget, &child_area);
	}
    }
}

static gint
gtk_vlayout_expose (GtkWidget      *widget, 
		   GdkEventExpose   *event)
{
    GtkVLayout *vlayout;
    GtkVLayoutChild *child;
    GList *children;
    GdkEventExpose child_event;

    g_return_val_if_fail (widget != NULL, FALSE);
    g_return_val_if_fail (GTK_IS_VLAYOUT(widget), FALSE);
    g_return_val_if_fail (event != NULL, FALSE);

    if (GTK_WIDGET_DRAWABLE (widget)) {
	vlayout = GTK_VLAYOUT (widget);

	child_event = *event;

	for (children = vlayout->children; children; children = children->next) {
	    child = children->data;

	    if (GTK_WIDGET_NO_WINDOW (child->widget) &&
		    gtk_widget_intersect (child->widget, &event->area, &child_event.area))
		gtk_widget_event (child->widget, (GdkEvent*) &child_event);
	}
    }
}
